@@ -6,6 +6,7 @@ import os
 import re
 import json
 import time
+from functools import lru_cache
 from queue import Queue
 from typing import List, Dict, Optional, Deque
 
@@ -14,6 +15,7 @@ from botocore.exceptions import ClientError, EndpointConnectionError
 from botocore.config import Config
 
 from dragoneye.cloud_scanner.aws.aws_scan_settings import AwsCloudScanSettings
+from dragoneye.utils.boto_backoff import rate_limiter
 from dragoneye.cloud_scanner.base_cloud_scanner import BaseCloudScanner
 from dragoneye.utils.app_logger import logger
 from dragoneye.utils.misc_utils import get_dynamic_values_from_files, custom_serializer, make_directory, init_directory, load_yaml, snakecase, \
@@ -43,6 +45,8 @@ class AwsScanner(BaseCloudScanner):
         if self.default_region is None:
             raise ValueError('Default region cannot be empty. '
                              'You must specify the default region or set the AWS_DEFAULT_REGION environment variable')
+        self.handler_config = Config(retries={'max_attempts': self.settings.max_attempts, 'mode': 'standard'},
+                                     max_pool_connections=self.settings.max_pool_connections)
         logging.getLogger("botocore").setLevel(logging.WARN)
 
     @elapsed_time('Scanning AWS live environment took {} seconds')
@@ -171,17 +175,19 @@ class AwsScanner(BaseCloudScanner):
             "region": region
         }
 
+        params_string = '' if not parameters else ', '.join(f'{k}={v}' for k, v in parameters.items())
+        function_msg = f'{call_summary["service"]}.{call_summary["action"]}({params_string})'
+        logger.info(f'Invoking {function_msg}')
         data = AwsScanner._get_data(output_file, handler, method_to_call, parameters, checks, call_summary)
         AwsScanner._remove_unused_values(data)
         AwsScanner._save_results_to_file(output_file, data)
 
-        logger.info("finished call for {}".format(output_file))
+        logger.info(f'Results from {function_msg} were saved to {output_file}')
         summary.put_nowait(call_summary)
 
     @staticmethod
     def _get_data(output_file, handler, method_to_call, parameters, checks, call_summary):
-        logger.info("  Making call for {}".format(output_file))
-        data: dict = {}
+        data = None
         try:
             for retries in range(MAX_RETRIES):
                 data = AwsScanner._call_boto_function(output_file, handler, method_to_call, parameters)
@@ -273,6 +279,10 @@ class AwsScanner(BaseCloudScanner):
                     "NoSuchAccessPointPolicy" in str(ex)
             ):
                 logger.warning(f"  - {str(ex)}")
+            elif (
+                    "PolicyNotFound" in str(ex)
+            ):
+                logger.warning(f"  - {str(ex)}")
             else:
                 logger.warning(f"ClientError {ex}")
                 call_summary["exception"] = ex
@@ -280,12 +290,16 @@ class AwsScanner(BaseCloudScanner):
             logger.warning("EndpointConnectionError: {}".format(ex))
             call_summary["exception"] = ex
         except Exception as ex:
-            logger.warning("Exception: {}".format(ex))
-            call_summary["exception"] = ex
+            if "Parameter validation failed" in str(ex):
+                logger.warning(f"  - {str(ex)}")
+            else:
+                logger.warning("Exception: {}".format(ex))
+                call_summary["exception"] = ex
 
         return data
 
     @staticmethod
+    @rate_limiter()
     def _call_boto_function(output_file, handler, method_to_call, parameters):
         data = {}
         if handler.can_paginate(method_to_call):
@@ -335,19 +349,24 @@ class AwsScanner(BaseCloudScanner):
         region = copy.deepcopy(region)
         runner = copy.deepcopy(runner)
         region_name = region["RegionName"]
-        logger.info(
-            "* Getting {}:{}:{} info".format(region["RegionName"], runner["Service"], runner["Request"])
-        )
 
-        if not self._should_run_command_on_region(runner, region):
-            return
+        client_region = region_name
+        if runner['Service'] == 'globalaccelerator':
+            # globalaccelerator only has api endpoint in us-west-2
+            if region_name == self.default_region:
+                client_region = 'us-west-2'
+            else:
+                return
+        else:
+            if not self._should_run_command_on_region(runner, region):
+                return
+
         handler = self.session.client(
-            runner["Service"], region_name=region["RegionName"],
-            config=Config(retries={'max_attempts': self.settings.max_attempts, 'mode': 'standard'},
-                          max_pool_connections=self.settings.max_pool_connections)
+            runner["Service"], region_name=client_region,
+            config=self.handler_config
         )
 
-        filepath = os.path.join(account_dir, region["RegionName"], f'{runner["Service"]}-{runner["Request"]}')
+        filepath = os.path.join(account_dir, region_name, f'{runner["Service"]}-{runner["Request"]}')
 
         method_to_call = snakecase(runner["Request"])
         parameter_keys = set()
@@ -388,8 +407,7 @@ class AwsScanner(BaseCloudScanner):
 
         deque_tasks: Deque[List[ThreadedFunctionData]] = collections.deque()
         deque_tasks.append(tasks)
-        max_workers = 10 if runner['Service'] == 'lambda' else 20
-        execute_parallel_functions_in_threads(deque_tasks, max_workers, self.settings.command_timeout)
+        execute_parallel_functions_in_threads(deque_tasks, 20, self.settings.command_timeout)
 
     def _scan_region_data(self, region, account_dir, summary: Queue):
         dependable_commands = []
@@ -506,7 +524,7 @@ class AwsScanner(BaseCloudScanner):
         if runner["Service"] in self.universal_services:
             if region_dict["RegionName"] != self.default_region:
                 return False
-        elif runner["Service"] != 'eks' and region_dict["RegionName"] not in self.session.get_available_regions(runner["Service"]):
+        elif runner["Service"] != 'eks' and region_dict["RegionName"] not in self._get_available_regions(runner["Service"]):
             logger.info("Skipping region {}, as {} does not exist there"
                         .format(region_dict["RegionName"], runner["Service"]))
             return False
@@ -525,3 +543,7 @@ class AwsScanner(BaseCloudScanner):
                 group = parameter.get("Group", False)
                 param_groups = self._fill_dynamic_params(param_groups, name, value, group, account_dir, region)
         return param_groups
+
+    @lru_cache(maxsize=None)
+    def _get_available_regions(self, service: str):
+        return self.session.get_available_regions(service)
